import inspect
from dataclasses import dataclass, field
from numbers import Number
from typing import Any, Callable, Optional, Union

import numpy as np
import torch
from torch.optim import Optimizer

from agilerl.protocols import EvolvableAlgorithm
from agilerl.typing import NetworkType
from agilerl.utils.llm_utils import DummyOptimizer


@dataclass
class NetworkConfig:
    """Dataclass for storing the configuration of an evolvable network
    within an `EvolvableAlgorithm`.

    :param name: The name of the attribute where the network is stored.
    :type name: str
    :param eval_network: Whether the network is an evaluation network. This implies
    that the network is optimized during training. Default is False.

    :type eval_network: bool
    :param optimizer: The name of the optimizer that updates the network.
    :type optimizer: Optional[str]
    """

    name: str
    eval_network: bool = field(default=False)
    optimizer: Optional[str] = field(default=None)

    def __post_init__(self):
        if self.eval_network and self.optimizer is None:
            raise ValueError(
                "Evaluation network must have an optimizer associated with it."
            )


@dataclass
class OptimizerConfig:
    """Dataclass for storing the configuration of an optimizer within an `EvolvableAlgorithm`. Usually
    an optimizer will be used to update the parameters of a single evaluation network, but in some cases
    it may be used to update the parameters of multiple networks simultaneously. Here we provide the
    flexibility to specify such configurations for PyTorch optimizers.

    :param name: The name of the attribute where the optimizer is stored.
    :type name: str
    :param networks: The list of network attribute names that the optimizer will update.
    :type networks: list[str]
    :param lr: The learning rate of the optimizer.
    :type lr: str
    :param optimizer_cls: The optimizer class to be used.
    :type optimizer_cls: type[Optimizer]
    :param optimizer_kwargs: The keyword arguments to be passed to the optimizer.
    :type optimizer_kwargs: dict[str, Any]
    """

    name: str
    networks: Union[str, list[str]]
    lr: str
    optimizer_cls: Union[type[Optimizer], list[type[Optimizer]]]
    optimizer_kwargs: Union[dict[str, Any], list[dict[str, Any]]]

    def __post_init__(self):
        # Save optimizer_cls as string for serialization
        if isinstance(self.optimizer_cls, dict):
            self.optimizer_cls = {
                agent_id: cls.__name__ for agent_id, cls in self.optimizer_cls.items()
            }
        else:
            self.optimizer_cls = self.optimizer_cls.__name__

    def __eq__(self, other: "OptimizerConfig") -> bool:
        return self.name == other.name and self.networks == other.networks

    def get_optimizer_cls(self) -> Union[type[Optimizer], dict[str, type[Optimizer]]]:
        """Get the optimizer object/s from the stored configuration.

        :return: The optimizer object/s from the stored configuration.
        :rtype: Union[Optimizer, dict[str, Optimizer]]
        """
        name_to_cls = {
            "Adam": torch.optim.Adam,
            "AdamW": torch.optim.AdamW,
            "SGD": torch.optim.SGD,
            "RMSprop": torch.optim.RMSprop,
            "Adadelta": torch.optim.Adadelta,
            "Adagrad": torch.optim.Adagrad,
            "Adamax": torch.optim.Adamax,
            "ASGD": torch.optim.ASGD,
            "LBFGS": torch.optim.LBFGS,
            "Rprop": torch.optim.Rprop,
            "DummyOptimizer": DummyOptimizer,
        }
        if isinstance(self.optimizer_cls, dict):
            return {
                agent_id: name_to_cls[cls_name]
                for agent_id, cls_name in self.optimizer_cls.items()
            }

        return name_to_cls[self.optimizer_cls]


@dataclass
class RLParameter:
    """Dataclass for storing the configuration of a hyperparameter that will be mutated during
    training. The hyperparameter is defined by a range of values that it can take, and the
    shrink and grow factors that will be used to mutate the hyperparameter value.

    :param min: The minimum value that the hyperparameter can take. For numpy arrays, this will be broadcast.
    :type min: float
    :param max: The maximum value that the hyperparameter can take. For numpy arrays, this will be broadcast.
    :type max: float
    :param shrink_factor: The factor by which the hyperparameter will be shrunk during mutation. Default is 0.8.
    :type shrink_factor: float
    :param grow_factor: The factor by which the hyperparameter will be grown during mutation. Default is 1.2.
    :type grow_factor: float
    :param dtype: The data type of the hyperparameter. Default is float.
    :type dtype: Union[type[float], type[int], type[np.ndarray]]
    :param value: The current value of the hyperparameter. Default is None.
    :type value: Optional[Union[Number, np.ndarray]]
    """

    min: float
    max: float
    shrink_factor: float = 0.8
    grow_factor: float = 1.2
    dtype: Union[type[float], type[int], type[np.ndarray]] = float
    value: Optional[Union[Number, np.ndarray]] = field(default=None, init=False)

    def mutate(self) -> Union[Number, np.ndarray]:
        """Mutate the hyperparameter value by either growing or shrinking it.

        For scalar values (int/float), the mutation applies the grow/shrink factor uniformly.
        For numpy arrays, the mutation is applied element-wise, with proper broadcasting
        of min/max constraints and preservation of the original array's dtype.

        :return: The mutated hyperparameter value.
        :rtype: Union[Number, np.ndarray]
        """
        assert self.value is not None, "Hyperparameter value is not set"

        # Equal probability of growing or shrinking
        if torch.rand(1).item() < 0.5:
            # Shrinking
            if isinstance(self.value, np.ndarray):
                new_value = np.where(
                    self.value * self.shrink_factor > self.min,
                    self.value * self.shrink_factor,
                    self.min,
                )
            else:
                if self.value * self.shrink_factor > self.min:
                    new_value = self.value * self.shrink_factor
                else:
                    new_value = self.min
        else:
            # Growing
            if isinstance(self.value, np.ndarray):
                new_value = np.where(
                    self.value * self.grow_factor < self.max,
                    self.value * self.grow_factor,
                    self.max,
                )
            else:
                if self.value * self.grow_factor < self.max:
                    new_value = self.value * self.grow_factor
                else:
                    new_value = self.max

        # Clip the new value to the min and max
        if isinstance(new_value, np.ndarray):
            new_value = np.clip(new_value, self.min, self.max)
        else:
            new_value = min(max(new_value, self.min), self.max)

        # Cast the new value to the correct dtype
        if isinstance(new_value, np.ndarray):
            # Preserve the original array's dtype
            new_value = new_value.astype(self.value.dtype)
            self.value = new_value
        else:
            self.value = self.dtype(new_value)

        return self.value


class HyperparameterConfig:
    """Stores the RL hyperparameters that will be mutated during training. For each
    hyperparameter, we store the name of the attribute where the hyperparameter is
    stored, and the range of values that the hyperparameter can take."""

    def __init__(self, **kwargs: dict[str, RLParameter]):
        self.config = kwargs
        for key, value in kwargs.items():
            if not isinstance(value, RLParameter):
                raise ValueError(
                    "Expected RLParameter object for hyperparameter configuration."
                )

            setattr(self, key, value)

    def __repr__(self) -> str:
        return (
            "HyperparameterConfig(\n"
            + "\n".join([f"{key}: {value}" for key, value in self.config.items()])
            + "\n)"
        )

    def __bool__(self) -> bool:
        """Returns False if the config is empty, True otherwise.

        :return: Whether the config contains any hyperparameters
        :rtype: bool
        """
        return bool(self.config)

    def __eq__(self, other: "HyperparameterConfig") -> bool:
        return set(self.names()) == set(other.names())

    def __iter__(self):
        return iter(self.config)

    def __getitem__(self, key: str) -> RLParameter:
        return self.config[key]

    def names(self) -> list[str]:
        return list(self.config.keys())

    def items(self) -> dict[str, Any]:
        return self.config.items()

    def sample(self) -> tuple[str, RLParameter]:
        """Sample a hyperparameter from the configuration.

        :return: The name of the hyperparameter and its configuration.
        :rtype: tuple[str, RLParameter]
        """
        key = torch.randperm(len(self.config))[0]
        return list(self.config.keys())[key], list(self.config.values())[key]


@dataclass
class NetworkGroup:
    """Dataclass for storing a group of networks. This consists of an evaluation network (i.e.
    a network that is optimized during training) and, optionally, some other networks that
    share parameters with the evaluation network (e.g. the target network in DQN). If the
    networks are passed as an agilerl.modules.base.ModuleDict, we assume that the networks
    are part of a multiagent setting.

    :param eval_network: The evaluation network.
    :type eval_network: NetworkType
    :param shared_networks: The list of shared networks.
    :type shared_networks: Optional[NetworkType]
    :param policy: Whether the network is a policy (e.g. the network used to get the actions
        of the agent). There must be one network group in an algorithm which sets this to True.
        Default is False.
    :type policy: bool
    """

    eval_network: NetworkType
    shared_networks: Optional[NetworkType] = field(default=None)
    policy: bool = field(default=False)

    def __post_init__(self):
        # Check that the shared networks are of the same type as the eval network
        if self.shared_networks is not None:
            eval_cls = type(self.eval_network)
            if isinstance(self.shared_networks, list):
                assert all(isinstance(net, eval_cls) for net in self.shared_networks), (
                    f"Expected a list of {eval_cls.__name__} objects for the "
                    f"shared argument in the network group. Found {type(self.shared_networks[0])}."
                )
            else:
                assert isinstance(self.shared_networks, eval_cls), (
                    f"Expected a {eval_cls.__name__} object for the "
                    f"shared argument in the network group. Found {type(self.shared_networks[0])}."
                )

        # Identify the names of the attributes where the networks are stored
        container = self._infer_parent_container()
        self.eval_network = self._infer_attribute_names(container, self.eval_network)[0]
        if self.shared_networks is not None:
            shared = (
                self.shared_networks
                if isinstance(self.shared_networks, list)
                else [self.shared_networks]
            )
            self.shared_networks = self._infer_attribute_names(container, shared)

    def __hash__(self) -> int:
        return hash((self.eval_network, self.shared_networks, self.policy))

    def _infer_parent_container(self) -> EvolvableAlgorithm:
        """
        Infer the parent container dynamically using the stack frame.

        :return: The parent container object
        :rtype: EvolvableAlgorithm
        """
        # NOTE: Here the assumption is that NetworkGroup is used inside the __init__
        # method of the implemented algorithm, such that we can access the defined locals
        # and extract the corresponding attribute names to the passed networks.
        current_frame = inspect.currentframe()
        return current_frame.f_back.f_back.f_back.f_locals["self"]

    def _infer_attribute_names(
        self, container: object, objects: Union[object, list[object]]
    ) -> list[str]:
        """
        Infer attribute names of the networks being optimized.

        :param container: The container object to inspect.
        :type container: object
        :param objects: The objects to match.
        :type objects: Union[object, list[object]]

        :return: List of attribute names for the networks
        :rtype: list[str]
        """

        def _match_condition(attr_value: Any) -> bool:
            if isinstance(objects, list):
                return any(id(attr_value) == id(obj) for obj in objects)
            else:
                return id(attr_value) == id(objects)

        return [
            attr_name
            for attr_name, attr_value in vars(container).items()
            if _match_condition(attr_value)
        ]


def make_network_group(
    eval_network: str,
    shared_networks: Optional[Union[str, list[str]]],
    policy: bool = False,
) -> NetworkGroup:
    """Make a network group from a given eval network and, optionally, some network/s that
    share parameters with the eval network.

    :param eval_network: The evaluation network.
    :type eval_network: str
    :param shared_networks: The list of shared networks.
    :type shared_networks: Optional[Union[str, list[str]]]
    :param policy: Whether the network is a policy (e.g. the network used to get the actions
    of the agent). There must be one network group in an algorithm which sets this to True.
    Default is False.
    :type policy: bool

    :return: NetworkGroup object with the passed configuration.
    :rtype: NetworkGroup
    """
    return NetworkGroup(
        eval_network=eval_network, shared_networks=shared_networks, policy=policy
    )


@dataclass
class MutationRegistry:
    """Registry to keep track of the components of an algorithms that may evolve during training
    in a structured way to be interpreted by a :class:`Mutations <agilerl.hpo.mutations.Mutations>` object
    when performing evolutionary hyperparameter optimization. This includes:

    1. The hyperparameter configuration of the algorithm.
    2. The network groups of the algorithm.
    3. The optimizers of the algorithm.
    4. The mutation hooks of the algorithm (i.e. functions that are called after a mutation is performed).

    :param hp_config: The hyperparameter configuration of the algorithm.
    :type hp_config: HyperparameterConfig
    """

    hp_config: Optional[HyperparameterConfig] = field(default=None)

    def __post_init__(self):
        self.groups: list[NetworkGroup] = []
        self.optimizers: list[OptimizerConfig] = []
        self.hooks: list[Callable] = []

        if self.hp_config is None:
            self.hp_config = HyperparameterConfig()

    def __repr__(self) -> str:
        groups_str = "\n".join(
            [
                f"Eval: '{group.eval_network}', Shared: {group.shared_networks}"
                for group in self.groups
            ]
        )
        optimizers_str = "\n".join(
            [
                f"{opt.optimizer_cls}: '{opt.name}', Networks: {opt.networks}"
                for opt in self.optimizers
            ]
        )
        return f"Network Groups:\n{groups_str}\n\nOptimizers:\n{optimizers_str}"

    def __eq__(self, other: Optional["MutationRegistry"]) -> bool:
        """Check if two MutationRegistry objects are equal. This involves checking
        that the network groups and optimizer configurations are the same.

        :param other: The other MutationRegistry object to compare with.
        :type other: Optional[MutationRegistry]

        :return: True if the two MutationRegistry objects are equal, False otherwise.
        :rtype: bool
        """
        return self.groups == other.groups and self.optimizers == other.optimizers

    @property
    def optimizer_networks(self) -> dict[str, list[str]]:
        """Get a dictionary of optimizer names and the network attribute names that they update.

        :return: A dictionary of optimizer names and the network attribute names that they update.
        :rtype: dict[str, list[str]]
        """
        return {config.name: config.networks for config in self.optimizers}

    def policy(self, return_group: bool = False) -> Optional[Union[str, NetworkGroup]]:
        """Get the name of the policy network in the registry.

        :param return_group: Whether to return the network group instead of just the name.
        :type return_group: bool

        :return: The name of the policy network in the registry.
        :rtype: Optional[Union[str, NetworkGroup]]
        """
        for group in self.groups:
            if group.policy:
                return group.eval_network if not return_group else group
        return

    def all_registered(self) -> list[str]:
        """Returns all of the members in the registry.

        :return: A list of all the members in the registry.
        :rtype: list[str]
        """
        all_registered = {group.eval_network for group in self.groups}
        all_registered.update(
            shared
            for group in self.groups
            if group.shared_networks is not None
            for shared in (
                group.shared_networks
                if isinstance(group.shared_networks, list)
                else [group.shared_networks]
            )
        )
        all_registered.update(opt.name for opt in self.optimizers)
        return all_registered

    def networks(self) -> list[NetworkConfig]:
        """Get a list of network configurations in the registry.

        :return: A list of network configurations in the registry. This includes the evaluation
        and shared networks.
        :rtype: list[NetworkConfig]
        """
        # Match with optimizers (only eval networks can have optimizers by definition)
        optimizer_eval = {}
        for opt_name, nets in self.optimizer_networks.items():
            for net in nets:
                optimizer_eval[net] = opt_name

        # Fetch evaluation and shared networks
        eval_networks = [
            NetworkConfig(
                name=group.eval_network,
                eval_network=True,
                optimizer=optimizer_eval.get(group.eval_network),
            )
            for group in self.groups
        ]
        shared_networks = [
            NetworkConfig(name=shared, eval_network=False)
            for group in self.groups
            if group.shared_networks is not None
            for shared in (
                group.shared_networks
                if isinstance(group.shared_networks, list)
                else [group.shared_networks]
            )
        ]

        return eval_networks + shared_networks

    def register_group(self, group: NetworkGroup) -> None:
        """Register a network configuration in the registry.

        :param group: The network group to be registered.
        :type group: NetworkGroup
        """
        self.groups.append(group)

    def register_optimizer(self, optimizer: OptimizerConfig) -> None:
        """Register an optimizer configuration in the registry.

        :param optimizer: The optimizer configuration to be registered.
        :type optimizer: OptimizerConfig
        """
        self.optimizers.append(optimizer)

    def register_hook(self, hook: Callable) -> None:
        """Register a hook in the registry as its name. This is used to store the names of the
        mutation hooks that will be applied after a mutation is performed.

        :param hook: The hook to be registered.
        :type hook: Callable
        """
        self.hooks.append(hook.__name__)



import random

import gymnasium as gym
import numpy as np
import torch
from gymnasium import spaces
from tqdm import trange

from agilerl.components.data import Transition


class ConstantRewardEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Discrete(1)
        self.action_space = spaces.Discrete(1)
        self.sample_obs = [np.zeros((1, 1))]
        self.q_values = [[1.0]]  # Correct Q values to learn, s x a table
        self.v_values = [[1.0]]  # Correct V values to learn, s table
        self.policy_values = [None]  # Correct policy to learn

    def step(self, action):
        observation = 0
        reward = 1  # Constant reward of 1
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        observation = 0
        info = {}
        return observation, info


class ConstantRewardImageEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Box(0.0, 0.0, (1, 3, 3))
        self.action_space = spaces.Discrete(1)
        self.sample_obs = [np.zeros((1, 1, 3, 3))]
        self.q_values = [[1.0]]  # Correct Q values to learn, s x a table
        self.v_values = [[1.0]]  # Correct V values to learn, s table
        self.policy_values = [None]  # Correct policy to learn

    def step(self, action):
        observation = np.zeros((1, 3, 3))
        reward = 1  # Constant reward of 1
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        observation = np.zeros((1, 3, 3))
        info = {}
        return observation, info


class ConstantRewardDictEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Dict(
            {"discrete": spaces.Discrete(1), "box": spaces.Box(0.0, 0.0, (1, 3, 3))}
        )
        self.action_space = spaces.Discrete(1)
        self.sample_obs = [{"discrete": np.array([[0]]), "box": np.zeros((1, 1, 3, 3))}]
        self.q_values = [[1.0]]  # Correct Q values to learn, s x a table
        self.v_values = [[1.0]]  # Correct V values to learn, s table
        self.policy_values = [None]  # Correct policy to learn

    def step(self, action):
        observation = {"discrete": 0, "box": np.zeros((1, 3, 3))}
        reward = 1  # Constant reward of 1
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        observation = {"discrete": 0, "box": np.zeros((1, 3, 3))}
        info = {}
        return observation, info


class ConstantRewardContActionsEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Discrete(1)
        self.action_space = spaces.Box(0.0, 1.0, (1,))
        self.sample_obs = [np.zeros((1, 1))]
        self.sample_actions = [[[1.0]]]
        self.q_values = [[1.0]]  # Correct Q values to learn, s x a table
        self.v_values = [[1.0]]  # Correct V values to learn, s table
        self.policy_values = [None]  # Correct policy to learn

    def step(self, action):
        observation = 0
        reward = 1  # Constant reward of 1
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        observation = 0
        info = {}
        return observation, info


class ConstantRewardContActionsImageEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Box(0.0, 0.0, (1, 3, 3))
        self.action_space = spaces.Box(0.0, 1.0, (1,))
        self.sample_obs = [np.zeros((1, 1, 3, 3))]
        self.sample_actions = [[[1.0]]]
        self.q_values = [[1.0]]  # Correct Q values to learn, s x a table
        self.v_values = [[1.0]]  # Correct V value to learn, s table
        self.policy_values = [None]  # Correct policy to learn

    def step(self, action):
        observation = np.zeros((1, 3, 3))
        reward = 1  # Constant reward of 1
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        observation = np.zeros((1, 3, 3))
        info = {}
        return observation, info


class ConstantRewardContActionsDictEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Dict(
            {"discrete": spaces.Discrete(1), "box": spaces.Box(0.0, 0.0, (1, 3, 3))}
        )
        self.action_space = spaces.Box(0.0, 1.0, (1,))
        self.sample_obs = [{"discrete": np.array([[0]]), "box": np.zeros((1, 1, 3, 3))}]
        self.sample_actions = [np.array([[1.0]])]
        self.q_values = [[1.0]]  # Correct Q values to learn, s x a table
        self.v_values = [[1.0]]  # Correct V values to learn, s table
        self.policy_values = [None]  # Correct policy to learn

    def step(self, action):
        observation = {"discrete": 0, "box": np.zeros((1, 3, 3))}
        reward = 1  # Constant reward of 1
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        observation = {"discrete": 0, "box": np.zeros((1, 3, 3))}
        info = {}
        return observation, info


class ObsDependentRewardEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Discrete(2)
        self.action_space = spaces.Discrete(1)
        self.last_obs = 1
        self.sample_obs = [np.array([[1, 0]]), np.array([[0, 1]])]
        self.q_values = [[-1.0], [1.0]]  # Correct Q values to learn, s x a table
        self.v_values = [[-1.0], [1.0]]  # Correct V values to learn, s table

    def step(self, action):
        observation = self.last_obs
        reward = -1 if self.last_obs == 0 else 1  # Reward depends on observation
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = random.choice([0, 1])
        info = {}
        return self.last_obs, info


class ObsDependentRewardImageEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Box(0.0, 1.0, (1, 3, 3))
        self.action_space = spaces.Discrete(1)
        self.last_obs = np.ones((1, 3, 3))
        self.sample_obs = [np.zeros((1, 1, 3, 3)), np.ones((1, 1, 3, 3))]
        self.q_values = [[-1.0], [1.0]]  # Correct Q values to learn, s x a table
        self.v_values = [[-1.0], [1.0]]  # Correct V values to learn, s table

    def step(self, action):
        observation = self.last_obs
        reward = (
            -1 if np.mean(self.last_obs) == 0.0 else 1
        )  # Reward depends on observation
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = random.choice([np.zeros((1, 3, 3)), np.ones((1, 3, 3))])
        info = {}
        return self.last_obs, info


class ObsDependentRewardDictEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Dict(
            {"discrete": spaces.Discrete(1), "box": spaces.Box(0.0, 1.0, (1, 3, 3))}
        )
        self.action_space = spaces.Discrete(1)
        self.last_obs = {"discrete": 1, "box": np.ones((1, 3, 3))}
        self.sample_obs = [
            {
                "discrete": [np.array([[0]]), np.array([[1]])],
                "box": [np.zeros((1, 1, 3, 3)), np.ones((1, 1, 3, 3))],
            }
        ]
        self.q_values = [[-1.0], [1.0]]  # Correct Q values to learn, s x a table
        self.v_values = [[-1.0], [1.0]]  # Correct V values to learn, s table

    def step(self, action):
        observation = {
            "discrete": self.last_obs["discrete"],
            "box": self.last_obs["box"],
        }
        reward = (
            -1 if np.mean(self.last_obs["box"]) != self.last_obs["discrete"] else 1
        )  # Reward depends on observation
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = {
            "discrete": random.choice([0, 1]),
            "box": random.choice([np.zeros((1, 3, 3)), np.ones((1, 3, 3))]),
        }
        info = {}
        return self.last_obs, info


class ObsDependentRewardContActionsEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Discrete(2)
        self.action_space = spaces.Box(0.0, 1.0, (1,))
        self.last_obs = 1
        self.sample_obs = [np.array([[1, 0]]), np.array([[0, 1]])]
        self.sample_actions = [[[1.0]], [[1.0]]]
        self.q_values = [[-1.0], [1.0]]  # Correct Q values to learn, s x a table
        self.policy_values = [None]  # Correct policy to learn
        self.v_values = [[-1.0], [1.0]]  # Correct V values to learn, s table

    def step(self, action):
        observation = self.last_obs
        reward = -1 if self.last_obs == 0 else 1  # Reward depends on observation
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = random.choice([0, 1])
        info = {}
        return self.last_obs, info


class ObsDependentRewardContActionsImageEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Box(0.0, 1.0, (1, 3, 3))
        self.action_space = spaces.Box(0.0, 1.0, (1,))
        self.last_obs = np.ones((1, 3, 3))
        self.sample_obs = [np.zeros((1, 1, 3, 3)), np.ones((1, 1, 3, 3))]
        self.sample_actions = [[[1.0]], [[1.0]]]
        self.q_values = [[-1.0], [1.0]]  # Correct Q values to learn, s x a table
        self.v_values = [[-1.0], [1.0]]  # Correct V values to learn, s table
        self.policy_values = [None]  # Correct policy to learn

    def step(self, action):
        observation = self.last_obs
        reward = (
            -1 if np.mean(self.last_obs) == 0.0 else 1
        )  # Reward depends on observationspaces.Box(0.0, 1.0
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = random.choice([np.zeros((1, 3, 3)), np.ones((1, 3, 3))])
        info = {}
        return self.last_obs, info


class ObsDependentRewardContActionsDictEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Dict(
            {"discrete": spaces.Discrete(1), "box": spaces.Box(0.0, 1.0, (1, 3, 3))}
        )
        self.action_space = spaces.Box(0.0, 1.0, (1,))
        self.last_obs = {"discrete": 1, "box": np.ones((1, 3, 3))}
        self.sample_obs = [
            {
                "discrete": [np.array([[0]]), np.array([[1]])],
                "box": [np.zeros((1, 1, 3, 3)), np.ones((1, 1, 3, 3))],
            }
        ]
        self.sample_actions = [[[1.0]], [[1.0]]]
        self.q_values = [[-1.0], [1.0]]  # Correct Q values to learn, s x a table
        self.v_values = [[-1.0], [1.0]]  # Correct V values to learn, s table
        self.policy_values = [None]  # Correct policy to learn

    def step(self, action):
        observation = {
            "discrete": self.last_obs["discrete"],
            "box": self.last_obs["box"],
        }
        reward = (
            -1 if np.mean(self.last_obs["box"]) != self.last_obs["discrete"] else 1
        )  # Reward depends on observation
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = {
            "discrete": random.choice([0, 1]),
            "box": random.choice([np.zeros((1, 3, 3)), np.ones((1, 3, 3))]),
        }
        info = {}
        return self.last_obs, info


class DiscountedRewardEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Discrete(2)
        self.action_space = spaces.Discrete(1)
        self.last_obs = 0
        self.sample_obs = [np.array([[1, 0]]), np.array([[0, 1]])]
        self.q_values = [[0.99], [1.0]]  # Correct Q values to learn, s x a table
        self.v_values = [[0.99], [1.0]]  # Correct V values to learn, s table

    def step(self, action):
        observation = 1
        reward = self.last_obs  # Reward depends on observation
        terminated = self.last_obs  # Terminate after second step
        truncated = False
        info = {}
        self.last_obs = 1
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = 0
        info = {}
        return self.last_obs, info


class DiscountedRewardImageEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Box(0.0, 1.0, (1, 3, 3))
        self.action_space = spaces.Discrete(1)
        self.last_obs = np.zeros((1, 3, 3))
        self.sample_obs = [np.zeros((1, 1, 3, 3)), np.ones((1, 1, 3, 3))]
        self.q_values = [[0.99], [1.0]]  # Correct Q values to learn, s x a table
        self.v_values = [[0.99], [1.0]]  # Correct V values to learn, s table

    def step(self, action):
        observation = np.ones((1, 3, 3))
        reward = np.mean(self.last_obs)  # Reward depends on observation
        terminated = int(np.mean(self.last_obs))  # Terminate after second step
        truncated = False
        info = {}
        self.last_obs = np.ones((1, 3, 3))
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = np.zeros((1, 3, 3))
        info = {}
        return self.last_obs, info


class DiscountedRewardDictEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Dict(
            {"discrete": spaces.Discrete(1), "box": spaces.Box(0.0, 1.0, (1, 3, 3))}
        )
        self.action_space = spaces.Discrete(1)
        self.last_obs = {"discrete": 0, "box": np.zeros((1, 3, 3))}
        self.sample_obs = [
            {
                "discrete": [np.array([[0]]), np.array([[1]])],
                "box": [np.zeros((1, 1, 3, 3)), np.ones((1, 1, 3, 3))],
            }
        ]
        self.q_values = [[0.99], [1.0]]  # Correct Q values to learn, s x a table
        self.v_values = [[0.99], [1.0]]  # Correct V values to learn, s table

    def step(self, action):
        observation = {"discrete": 1, "box": np.ones((1, 3, 3))}
        reward = (
            np.mean(self.last_obs["box"]) + self.last_obs["discrete"]
        )  # Reward depends on observation
        terminated = int(np.mean(self.last_obs["box"]))  # Terminate after second step
        truncated = False
        info = {}
        self.last_obs = {"discrete": 1, "box": np.ones((1, 3, 3))}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = {"discrete": 0, "box": np.zeros((1, 3, 3))}
        info = {}
        return self.last_obs, info


class DiscountedRewardContActionsEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Discrete(2)
        self.action_space = spaces.Box(0.0, 1.0, (1,))
        self.last_obs = 0
        self.sample_obs = [np.array([[1, 0]]), np.array([[0, 1]])]
        self.sample_actions = [[[1.0]], [[1.0]]]
        self.q_values = [[0.99], [1.0]]  # Correct Q values to learn, s x a table
        self.v_values = [[0.99], [1.0]]  # Correct V values to learn, s table
        self.policy_values = [None]  # Correct policy to learn

    def step(self, action):
        observation = 1
        reward = self.last_obs  # Reward depends on observation
        terminated = self.last_obs  # Terminate after second step
        truncated = False
        info = {}
        self.last_obs = 1
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = 0
        info = {}
        return self.last_obs, info


class DiscountedRewardContActionsImageEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Box(0.0, 1.0, (1, 3, 3))
        self.action_space = spaces.Box(0.0, 1.0, (1,))
        self.last_obs = np.zeros((1, 3, 3))
        self.sample_obs = [np.zeros((1, 1, 3, 3)), np.ones((1, 1, 3, 3))]
        self.sample_actions = [[[1.0]], [[1.0]]]
        self.q_values = [[0.99], [1.0]]  # Correct Q values to learn, s x a table
        self.v_values = [[0.99], [1.0]]  # Correct V values to learn, s table
        self.policy_values = [None]  # Correct policy to learn

    def step(self, action):
        observation = np.ones((1, 3, 3))
        reward = np.mean(self.last_obs)  # Reward depends on observation
        terminated = int(np.mean(self.last_obs))  # Terminate after second step
        truncated = False
        info = {}
        self.last_obs = np.ones((1, 3, 3))
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = np.zeros((1, 3, 3))
        info = {}
        return self.last_obs, info


class DiscountedRewardContActionsDictEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Dict(
            {"discrete": spaces.Discrete(1), "box": spaces.Box(0.0, 1.0, (1, 3, 3))}
        )
        self.action_space = spaces.Box(0.0, 1.0, (1,))
        self.last_obs = {"discrete": 0, "box": np.zeros((1, 3, 3))}
        self.sample_obs = [
            {"discrete": np.array([[0]]), "box": np.zeros((1, 1, 3, 3))},
            {"discrete": np.array([[1]]), "box": np.ones((1, 1, 3, 3))},
        ]
        self.sample_actions = self.sample_actions = [[[1.0]], [[1.0]]]
        self.q_values = [[0.99], [1.0]]  # Correct Q values to learn, s x a table
        self.v_values = [[0.99], [1.0]]  # Correct V values to learn, s table
        self.policy_values = [None]  # Correct policy to learn

    def step(self, action):
        observation = {"discrete": 1, "box": np.ones((1, 3, 3))}
        reward = (
            np.mean(self.last_obs["box"]) + self.last_obs["discrete"]
        )  # Reward depends on observation
        terminated = int(np.mean(self.last_obs["box"]))  # Terminate after second step
        truncated = False
        info = {}
        self.last_obs = {"discrete": 1, "box": np.ones((1, 3, 3))}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = {"discrete": 0, "box": np.zeros((1, 3, 3))}
        info = {}
        return self.last_obs, info


class FixedObsPolicyEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Discrete(1)
        self.action_space = spaces.Discrete(2)
        self.sample_obs = [np.array([[0]])]
        self.q_values = [[-1.0, 1.0]]  # Correct Q values to learn, s x a table
        self.v_values = [None]  # Correct V values to learn, s table
        self.policy_values = [[0.0, 1.0]]  # Correct policy to learn

    def step(self, action):
        if isinstance(action, (np.ndarray, list)):
            action = action[0]
        observation = 0
        reward = [-1, 1][action]  # Reward depends on action
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        observation = 0
        info = {}
        return observation, info


class FixedObsPolicyImageEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Box(0.0, 0.0, (1, 3, 3))
        self.action_space = spaces.Discrete(2)
        self.sample_obs = [np.zeros((1, 1, 3, 3))]
        self.q_values = [[-1.0, 1.0]]  # Correct Q values to learn, s x a table
        self.v_values = [None]  # Correct V values to learn, s table
        self.policy_values = [[0.0, 1.0]]  # Correct policy to learn

    def step(self, action):
        observation = np.zeros((1, 3, 3))
        if isinstance(action, (np.ndarray, list)):
            action = action[0]
        reward = [-1, 1][action]  # Reward depends on action
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        observation = np.zeros((1, 3, 3))
        info = {}
        return observation, info


class FixedObsPolicyDictEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Dict(
            {"discrete": spaces.Discrete(1), "box": spaces.Box(0.0, 0.0, (1, 3, 3))}
        )
        self.action_space = spaces.Discrete(2)
        self.sample_obs = [{"discrete": np.array([[0]]), "box": np.zeros((1, 1, 3, 3))}]
        self.q_values = [[-1.0, 1.0]]  # Correct Q values to learn, s x a table
        self.v_values = [None]  # Correct V values to learn, s table
        self.policy_values = [[0.0, 1.0]]  # Correct policy to learn

    def step(self, action):
        observation = {"discrete": 0, "box": np.zeros((1, 3, 3))}
        if isinstance(action, (np.ndarray, list)):
            action = action[0]
        reward = [-1, 1][action]  # Reward depends on action
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        observation = {"discrete": 0, "box": np.zeros((1, 3, 3))}
        info = {}
        return observation, info


class FixedObsPolicyContActionsEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Discrete(1)
        self.action_space = spaces.Box(0.0, 1.0, (1,))
        self.sample_obs = [np.array([[0]])]
        self.sample_actions = [np.array([[1.0]])]
        self.q_values = np.array([[0.0]])  # Correct Q values to learn, s x a table
        self.v_values = [None]  # Correct V values to learn, s table
        self.policy_values = [[1.0]]  # Correct policy to learn

    def step(self, action):
        observation = 0
        reward = -((1 - action[0]) ** 2)  # Reward depends on action
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        observation = 0
        info = {}
        return observation, info


class FixedObsPolicyContActionsImageEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Box(0.0, 1.0, (1, 3, 3))
        self.action_space = spaces.Box(0.0, 1.0, (1,))
        self.sample_obs = [np.zeros((1, 1, 3, 3))]
        self.sample_actions = [np.array([[1.0]])]
        self.q_values = np.array([[0.0]])  # Correct Q values to learn, s x a table
        self.v_values = [None]  # Correct V values to learn, s table
        self.policy_values = [[1.0]]  # Correct policy to learn

    def step(self, action):
        observation = np.zeros((1, 3, 3))
        reward = -((1 - action[0]) ** 2)  # Reward depends on action
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        observation = np.zeros((1, 3, 3))
        info = {}
        return observation, info


class FixedObsPolicyContActionsDictEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Dict(
            {"discrete": spaces.Discrete(1), "box": spaces.Box(0.0, 1.0, (1, 3, 3))}
        )
        self.action_space = spaces.Box(0.0, 1.0, (1,))
        self.sample_obs = [{"discrete": np.array([[0]]), "box": np.zeros((1, 1, 3, 3))}]
        self.sample_actions = [np.array([[1.0]])]
        self.q_values = np.array([[0.0]])  # Correct Q values to learn, s x a table
        self.v_values = [None]  # Correct V values to learn, s table
        self.policy_values = [[1.0]]  # Correct policy to learn

    def step(self, action):
        observation = {"discrete": 0, "box": np.zeros((1, 3, 3))}
        reward = -((1 - action[0]) ** 2)  # Reward depends on action
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        observation = {"discrete": 0, "box": np.zeros((1, 3, 3))}
        info = {}
        return observation, info


class PolicyEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Discrete(2)
        self.action_space = spaces.Discrete(2)
        self.last_obs = 0
        self.sample_obs = [np.array([[1, 0]]), np.array([[0, 1]])]
        self.q_values = [
            [1.0, -1.0],
            [-1.0, 1.0],
        ]  # Correct Q values to learn, s x a table
        self.v_values = [None]  # Correct V values to learn, s table

    def step(self, action):
        observation = self.last_obs
        reward = (
            1 if action == self.last_obs else -1
        )  # Reward depends on action in observation
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = random.choice([0, 1])
        info = {}
        return self.last_obs, info


class PolicyImageEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Box(0.0, 1.0, (1, 3, 3))
        self.action_space = spaces.Discrete(2)
        self.last_obs = np.ones((1, 3, 3))
        self.sample_obs = [np.zeros((1, 1, 3, 3)), np.ones((1, 1, 3, 3))]
        self.q_values = [
            [1.0, -1.0],
            [-1.0, 1.0],
        ]  # Correct Q values to learn, s x a table
        self.v_values = [None]  # Correct V values to learn, s table

    def step(self, action):
        observation = self.last_obs
        reward = (
            1 if action == int(np.mean(self.last_obs)) else -1
        )  # Reward depends on action in observation
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = random.choice([np.zeros((1, 3, 3)), np.ones((1, 3, 3))])
        info = {}
        return self.last_obs, info


class PolicyDictEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Dict(
            {"discrete": spaces.Discrete(2), "box": spaces.Box(0.0, 1.0, (1, 3, 3))}
        )
        self.action_space = spaces.Discrete(2)
        self.last_obs = {"discrete": 1, "box": np.ones((1, 3, 3))}
        self.sample_obs = [
            {"discrete": 0, "box": np.zeros((1, 3, 3))},
            {"discrete": 0, "box": np.ones((1, 3, 3))},
            {"discrete": 1, "box": np.zeros((1, 3, 3))},
            {"discrete": 1, "box": np.ones((1, 3, 3))},
        ]
        self.q_values = [
            [1.0, -1.0],  # discrete=0, box=0
            [-1.0, -1.0],  # discrete=0, box=1
            [-1.0, -1.0],  # discrete=1, box=0
            [-1.0, 1.0],  # discrete=1, box=1
        ]
        self.v_values = [None]

    def step(self, action):
        if isinstance(action, (np.ndarray, list)):
            action = action[0]
        observation = {
            "discrete": self.last_obs["discrete"],
            "box": self.last_obs["box"],
        }
        reward = (
            1
            if action == self.last_obs["discrete"]
            and action == int(np.mean(self.last_obs["box"]))
            else -1
        )
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = {
            "discrete": random.choice([0, 1]),
            "box": random.choice([np.zeros((1, 3, 3)), np.ones((1, 3, 3))]),
        }
        info = {}
        return self.last_obs, info


class PolicyContActionsEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Discrete(2)
        self.action_space = spaces.Box(0.0, 1.0, (2,))
        self.last_obs = 0
        self.sample_obs = [np.array([[1, 0]]), np.array([[0, 1]])]
        self.sample_actions = [np.array([[1.0, 0.0]]), np.array([[0.0, 1.0]])]
        self.q_values = [[0.0], [0.0]]  # Correct Q values to learn
        self.v_values = [None]  # Correct V values to learn, s table
        self.policy_values = [[1.0, 0.0], [0.0, 1.0]]  # Correct policy to learn

    def step(self, action):
        observation = self.last_obs
        if self.last_obs:  # last obs = 1, policy should be [0, 1]
            reward = -((0 - action[0]) ** 2) - (1 - action[1]) ** 2
        else:  # last obs = 0, policy should be [1, 0]
            reward = -((1 - action[0]) ** 2) - (0 - action[1]) ** 2
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = random.choice([0, 1])
        info = {}
        return self.last_obs, info


class PolicyContActionsImageEnvSimple(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Box(0.0, 1.0, (1, 3, 3))
        self.action_space = spaces.Box(0.0, 1.0, (1,))
        self.last_obs = np.zeros((1, 3, 3))
        self.sample_obs = [
            np.zeros((1, 1, 3, 3)),
            np.zeros((1, 1, 3, 3)),
            np.ones((1, 1, 3, 3)),
            np.ones((1, 1, 3, 3)),
        ]
        self.sample_actions = [
            np.array([[0.0]]),
            np.array([[1.0]]),
            np.array([[0.0]]),
            np.array([[1.0]]),
        ]
        self.q_values = [[0.0], [-1.0], [-1.0], [0.0]]  # Correct Q values to learn
        self.policy_values = [[0.0], [0.0], [1.0], [1.0]]  # Correct policy to learn
        self.v_values = [None]  # Correct V values to learn, s table

    def step(self, action):
        observation = self.last_obs
        if int(np.mean(self.last_obs)):  # last obs = 1, policy should be [1]
            reward = -((1 - action[0]) ** 2)
        else:  # last obs = 0, policy should be [0]
            reward = -((0 - action[0]) ** 2)
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        # self.last_obs = random.choice([np.zeros((1, 3, 3)), np.ones((1, 3, 3))])
        if int(np.mean(self.last_obs)):
            self.last_obs = np.zeros((1, 3, 3))
        else:
            self.last_obs = np.ones((1, 3, 3))

        info = {}
        return self.last_obs, info


class PolicyContActionsImageEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Box(0.0, 1.0, (1, 3, 3))
        self.action_space = spaces.Box(0.0, 1.0, (2,))
        self.last_obs = np.zeros((1, 3, 3))
        self.sample_obs = [np.zeros((1, 1, 3, 3)), np.ones((1, 1, 3, 3))]
        self.sample_actions = [np.array([[1.0, 0.0]]), np.array([[0.0, 1.0]])]
        self.q_values = [[0.0], [0.0]]  # Correct Q values to learn
        self.policy_values = [[1.0, 0.0], [0.0, 1.0]]  # Correct policy to learn
        self.v_values = [None]  # Correct V values to learn, s table

    def step(self, action):
        observation = self.last_obs
        if int(np.mean(self.last_obs)):  # last obs = 1, policy should be [0, 1]
            reward = -((0 - action[0]) ** 2) - (1 - action[1]) ** 2
        else:  # last obs = 0, policy should be [1, 0]
            reward = -((1 - action[0]) ** 2) - (0 - action[1]) ** 2
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = random.choice([np.zeros((1, 3, 3)), np.ones((1, 3, 3))])
        info = {}
        return self.last_obs, info


class PolicyContActionsDictEnv(gym.Env):
    def __init__(self):
        self.observation_space = spaces.Dict(
            {"discrete": spaces.Discrete(1), "box": spaces.Box(0.0, 1.0, (1, 3, 3))}
        )
        self.action_space = spaces.Box(0.0, 1.0, (2,))
        self.last_obs = {"discrete": 0, "box": np.zeros((1, 3, 3))}
        self.sample_obs = [
            {"discrete": np.array([[0]]), "box": np.zeros((1, 1, 3, 3))},
            {"discrete": np.array([[1]]), "box": np.ones((1, 1, 3, 3))},
        ]
        self.sample_actions = [[1.0, 0.0], 0.0, 1.0]
        self.q_values = [[0.0], [0.0]]  # Correct Q values to learn
        self.policy_values = [[1.0, 0.0], [0.0, 1.0]]  # Correct policy to learn
        self.v_values = [None]  # Correct V values to learn, s table

    def step(self, action):
        observation = {
            "discrete": self.last_obs["discrete"],
            "box": self.last_obs["box"],
        }
        if self.last_obs["discrete"] and int(
            np.mean(self.last_obs["box"])
        ):  # last obs = 1, policy should be [0, 1]
            reward = -((0 - action[0]) ** 2) - (1 - action[1]) ** 2
        else:  # last obs = 0, policy should be [1, 0]
            reward = -((1 - action[0]) ** 2) - (0 - action[1]) ** 2
        terminated = True
        truncated = False
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = {
            "discrete": random.choice([0, 1]),
            "box": random.choice([np.zeros((1, 3, 3)), np.ones((1, 3, 3))]),
        }
        info = {}
        return self.last_obs, info


def check_q_learning_with_probe_env(
    env, algo_class, algo_args, memory, learn_steps=10000, device="cpu"
):
    print(f"Probe environment: {type(env).__name__}")

    agent = algo_class(**algo_args, device=device)

    state, _ = env.reset()
    for _ in range(1000):
        if isinstance(state, dict):
            state = {k: np.expand_dims(v, 0) for k, v in state.items()}
        else:
            state = np.expand_dims(state, 0)
        action = agent.get_action(state, epsilon=1)
        next_state, reward, done, _, _ = env.step(action)
        transition = Transition(
            obs=state, action=action, reward=reward, next_obs=next_state, done=done
        ).to_tensordict()
        transition = transition.unsqueeze(0)
        transition.batch_size = [1]
        memory.add(transition)
        state = next_state
        if done:
            state, _ = env.reset()

    # Learn from experiences
    for i in trange(learn_steps):
        experiences = memory.sample(agent.batch_size)
        # Learn according to agent's RL algorithm
        loss = agent.learn(experiences)
        if i < 20:
            print("Loss = ", loss)

    for sample_obs, q_values in zip(env.sample_obs, env.q_values):
        predicted_q_values = agent.actor(sample_obs).detach().cpu().numpy()[0]
        print("q_values", q_values)
        print("predicted_q_values", predicted_q_values)
        # assert np.allclose(
        #     q_values, predicted_q_values, atol=0.1
        # ), f"{q_values} != {predicted_q_values}"


def check_policy_q_learning_with_probe_env(
    env, algo_class, algo_args, memory, learn_steps=10000, device="cpu"
):
    print(f"Probe environment: {type(env).__name__}")

    agent = algo_class(**algo_args, device=device)

    state, _ = env.reset()
    for _ in range(5000):
        action = (
            (agent.action_space.high - agent.action_space.low)
            * np.random.rand(1, agent.action_dim).astype("float32")
        ) + agent.action_space.low
        action = action[0]
        next_state, reward, done, _, _ = env.step(action)
        transition = Transition(
            obs=state, action=action, reward=reward, next_obs=next_state, done=done
        ).to_tensordict()
        transition = transition.unsqueeze(0)
        transition.batch_size = [1]
        memory.add(transition)
        state = next_state
        if done:
            state, _ = env.reset()

    # Learn from experiences
    for i in trange(learn_steps):
        experiences = memory.sample(agent.batch_size)
        # Learn according to agent's RL algorithm
        loss = agent.learn(experiences)
        if i < 20:
            print("Loss = ", loss)

    for sample_obs, sample_action, q_values, policy_values in zip(
        env.sample_obs, env.sample_actions, env.q_values, env.policy_values
    ):
        if isinstance(sample_obs, dict):
            state = {
                k: torch.tensor(v).float().to(device) for k, v in sample_obs.items()
            }
        else:
            state = torch.tensor(sample_obs).float().to(device)

        agent.critic.eval()
        agent.actor.eval()
        action = torch.tensor(sample_action).float().to(device)
        predicted_q_values = agent.critic(state, action).detach().cpu().numpy()[0]
        print("q", q_values, predicted_q_values)
        # assert np.allclose(
        #     q_values, predicted_q_values, atol=0.15
        # ), f"{q_values} != {predicted_q_values}"

        if policy_values is not None:
            predicted_policy_values = agent.actor(sample_obs).detach().cpu().numpy()[0]

            print("pol", policy_values, predicted_policy_values)
            # assert np.allclose(
            #     policy_values, predicted_policy_values, atol=0.2
            # ), f"{policy_values} != {predicted_policy_values}"


def check_policy_on_policy_with_probe_env(
    env,
    algo_class,
    algo_args,
    learn_steps=5000,
    device="cpu",
    discrete=True,
):
    print(f"Probe environment: {type(env).__name__}")

    agent = algo_class(**algo_args, device=device)

    for i in trange(learn_steps):
        state, _ = env.reset()
        states = []
        actions = []
        log_probs = []
        rewards = []
        dones = []
        values = []

        done = 0

        for j in range(200):
            if isinstance(state, dict):
                state = {k: np.expand_dims(v, 0) for k, v in state.items()}
            else:
                state = np.expand_dims(state, 0)

            action, log_prob, _, value = agent.get_action(state)

            action = action[0]
            log_prob = log_prob[0]
            value = value[0]
            next_state, reward, term, trunc, _ = env.step(action)
            next_done = np.logical_or(term, trunc).astype(np.int8)

            states.append(state)
            actions.append(action)
            log_probs.append(log_prob)
            rewards.append(reward)
            dones.append(done)
            values.append(value)

            state = next_state
            done = next_done

            if done:
                state, _ = env.reset()

        experiences = (
            states,
            actions,
            log_probs,
            rewards,
            dones,
            values,
            next_state,
            next_done,
        )
        loss = agent.learn(experiences)
        if i < 20:
            print("Loss = ", loss)

    for sample_obs, v_values, policy_values in zip(
        env.sample_obs, env.v_values, env.policy_values
    ):
        if isinstance(sample_obs, dict):
            state = {
                k: torch.tensor(v).float().to(device) for k, v in sample_obs.items()
            }
        else:
            state = torch.tensor(sample_obs).float().to(device)

        if v_values is not None:
            predicted_v_values = agent.critic(state).detach().cpu().numpy()[0]
            print("v", v_values, predicted_v_values)
            # assert np.allclose(
            #     v_values, predicted_v_values, atol=0.2
            # ), f"{v_values} != {predicted_v_values}"

        if policy_values is not None:
            # Assumes it is always a discrete action space
            _, _, _ = agent.actor(state)
            predicted_policy_values = (
                agent.actor.head_net.dist.distribution.probs.detach().cpu().numpy()
            )

            print("pol", policy_values, predicted_policy_values)
            # assert np.allclose(
            #     policy_values, predicted_policy_values, atol=0.2
            # ), f"{policy_values} != {predicted_policy_values}"
